#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "mwm.h"
/*
**  $Id: mwm.c,v 1.3 2003/03/13 13:58:42 scara Exp $
**
**  mwm.c   - A Modified Wu-Manber Style Multi-Pattern Matcher
**
** Copyright (C) 2002 Sourcefire,Inc
** Marc Norton
**
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** You can read more information and comments in the original snort source
** code..
*/

/*
*
*  Boyer-Moore-Horsepool for small pattern groups
*    
*/
struct hbm *hbm_prepx(struct hbm * p, unsigned char *pat, int m)
{
	int k;

	if (!m || !p)
		return 0;

	p->P = pat;
	p->M = m;

	/* Compute normal Boyer-Moore Bad Character Shift */
	for (k = 0; k < 256; k++)
		p->bcShift[k] = m;

	for (k = 0; k < m; k++)
		p->bcShift[pat[k]] = m - k - 1;

	return p;
}

struct hbm *hbm_prep(unsigned char *pat, int m)
{
	struct hbm *p;
	p = (struct hbm *) malloc(sizeof (struct hbm));
	if (!p)
		return 0;
	return hbm_prepx(p, pat, m);
}

static inline void case_normalise_str(unsigned char *s, int m)
{
	int i;
	for (i = 0; i < m; i++) {
		s[i] = tolower(s[i]);
	}
}

static inline int case_normalise(int s)
{
	return tolower(s);
}

int memcmpi(unsigned char *a, unsigned char *b, size_t n)
{
	unsigned char *e;

	for(e=a+n; a<e; a++,b++) {
		if ( *a!=case_normalise(*b) )
			return 1;
	}

	return 0;
}

/*
*   Boyer-Moore Horspool
*   Does NOT use Sentinel Byte(s)
*   Scan and Match Loops are unrolled and separated
*   Optimized for 1 byte patterns as well
*/
static INLINE unsigned char *
hbm_match(struct hbm * px, unsigned char *text, int n)
{
	unsigned char *pat, *t, *et, *q;
	int m1, k;
	strlen_t *bcShift;
	m1 = px->M - 1;
	pat = px->P;
	bcShift = px->bcShift;
	t = text + m1;
	et = text + n;

	/* Handle 1 Byte patterns - it's a faster loop */
	if (!m1) {
		for (; t < et; t++)
			if (case_normalise(*t) == *pat)
				return t;
		return 0;
	}

	/* Handle MultiByte Patterns */
	while (t < et) {
		unsigned char cur=case_normalise(*t);

		/* Scan Loop - Bad Character Shift */
		do {
			t += bcShift[cur];
			if (t >= et)
				return 0;;
			t += (k = bcShift[cur]);
			if (t >= et)
				return 0;
		} while (k);

		/* Unrolled Match Loop */
		k = m1;
		q = t - m1;
		while (k >= 4) {
			if (pat[k] != q[k])
				goto NoMatch;
			k--;
			if (pat[k] != q[k])
				goto NoMatch;
			k--;
			if (pat[k] != q[k])
				goto NoMatch;
			k--;
			if (pat[k] != q[k])
				goto NoMatch;
			k--;
		}

		/* Finish Match Loop */
		while (k >= 0) {
			if (pat[k] != q[k])
				goto NoMatch;
			k--;
		}

		/* If matched - return 1st char of pattern in text */
		return q;
	      NoMatch:
		/* Shift by 1, this replaces the good suffix shift */
		t++;
	}
	return 0;
}

void *mwmNew()
{
	struct mwm *p;
	
	if ( !(p=calloc(sizeof(*p), 1)) )
		return 0;

	/* XXX: if foo<msSmallest then blah, so needs
	 * to be a large signed int, hrm.
	 */
	p->msSmallest = 32000;

	return p;
}

/*
**
*/
void mwmFree(void *pv)
{
	struct mwm *p = (struct mwm *) pv;
	if (p) {
		if (p->msPatArray)
			free(p->msPatArray);
		if (p->msNumArray)
			free(p->msNumArray);
		if (p->msHash)
			free(p->msHash);
		if (p->msShift2)
			free(p->msShift2);
		free(p);
	}
}

/*
** mwmAddPatternEx::
**
** returns -1: max patterns exceeded
**          0: already present, uniqueness compiled in
**          1: added
*/
int mwmAddPatternEx(void *pv, unsigned char *P, int m, unsigned noCase,
		unsigned offset, unsigned depth, unsigned id, int iid)
{
	struct mwm *ps=pv;
	struct mwm_pat *plist=0;
	struct mwm_pat *p;
	
	if ( !(p=calloc(sizeof(*p), 1)) )
		return -1;

#ifdef REQUIRE_UNIQUE_PATTERNS
	for (plist = ps->plist; plist != NULL; plist = plist->next) {
		if (plist->psLen == m) {
			if (memcmp(P, plist->psPat, m) == 0)
				return 0;
		}
	}
#endif

	if (ps->plist) {
		for (plist = ps->plist; plist->next != NULL;
		     plist = plist->next) ;
		plist->next = p;
	}else{
		ps->plist = p;
	}

	/* Allocate and store the Pattern  'P' with NO CASE info */
	p->psPat = (unsigned char *) malloc(m);
	if (!p->psPat)
		return -1;
	memcpy(p->psPat, P, m);
	case_normalise_str(p->psPat, m);

	/* Allocate and store the Pattern  'P' with CASE info */
	p->psPatCase = (unsigned char *) malloc(m);
	if (!p->psPatCase)
		return -1;
	memcpy(p->psPatCase, P, m);
	p->psLen = m;
	p->psID = id;
	p->psIID = iid;
	p->psNoCase = noCase;
	p->psOffset = offset;
	p->psDepth = depth;
	ps->msNoCase += noCase;
	ps->msNumPatterns++;
	if (p->psLen < (unsigned) ps->msSmallest)
		ps->msSmallest = p->psLen;
	if (p->psLen > (unsigned) ps->msLargest)
		ps->msLargest = p->psLen;
	ps->msTotal += p->psLen;
	ps->msAvg = ps->msTotal / ps->msNumPatterns;
	return 1;
}

/*
** bcompare:: 
**
** Perform a Binary comparsion of 2 byte sequences of possibly 
** differing lengths.
**
** returns -1 a < b
**         +1 a > b
**          0 a = b
*/
static int bcompare(unsigned char *a, int alen, unsigned char *b, int blen)
{
	int stat;
	if (alen == blen) {
		return memcmp(a, b, alen);
	}

	else if (alen < blen) {
		if ((stat = memcmp(a, b, alen)) != 0)
			return stat;
		return -1;
	}

	else {
		if ((stat = memcmp(a, b, blen)) != 0)
			return stat;
		return +1;
	}
}

/*
** sortcmp::  qsort callback
*/
static int sortcmp(const void *e1, const void *e2)
{
	struct mwm_pat *r1 = (struct mwm_pat *) e1;
	struct mwm_pat *r2 = (struct mwm_pat *) e2;
	return bcompare(r1->psPat, r1->psLen, r2->psPat, r2->psLen);
}

/*
** HASH ROUTINE - used during pattern setup, but inline during searches
*/
static inline hash_t hash(unsigned char *T)
{
	return (unsigned short) (((*T)<<8) | *(T+1));
}

/*
** Build the hash table, and pattern groups
*/
static void
mwmPrepHashedPatternGroups(struct mwm * ps)
{
	unsigned sindex, hindex, ningroup;
	int i;

	/*
	   **  Allocate and Init 2+ byte pattern hash table 
	 */
	ps->msNumHashEntries = HASHTABLESIZE;
	ps->msHash =
	    (hash_t *) malloc(sizeof (hash_t) * ps->msNumHashEntries);
	if (!ps->msHash) {
		printf("No memory in mwmPrephashedPatternGroups()");
		exit(1);
	}

	/* Init Hash table to default value */
	for (i = 0; i < (int) ps->msNumHashEntries; i++) {
		ps->msHash[i] = (hash_t) - 1;
	}
	/* Initialize The One Byte Pattern Hash Table */
	for (i = 0; i < 256; i++) {
		ps->msHash1[i] = (hash_t) - 1;
	}

	/*
	   ** Add the patterns to the hash table
	 */
	for (i = 0; i < ps->msNumPatterns; i++) {
		ps->msPatArray[i].psIID = i;	/* internal id */
		if (ps->msPatArray[i].psLen > 1) {
			hindex = hash(ps->msPatArray[i].psPat);
			sindex = ps->msHash[hindex] = i;
			ningroup = 1;
			while ((++i < ps->msNumPatterns)
			       && (hindex == hash(ps->msPatArray[i].psPat)))
				ningroup++;
			ps->msNumArray[sindex] = ningroup;
			i--;
		}

		else if (ps->msPatArray[i].psLen == 1) {
			hindex = ps->msPatArray[i].psPat[0];
			sindex = ps->msHash1[hindex] = i;
			ningroup = 1;
			while ((++i < ps->msNumPatterns)
			       && (hindex == ps->msPatArray[i].psPat[0]))
				ningroup++;
			ps->msNumArray1[sindex] = ningroup;
			i--;
		}
	}
}

/*
*  Standard Bad Character Multi-Pattern Skip Table
*/
static void mwmPrepBadCharTable(struct mwm * ps)
{
	strlen_t i, k, m, cindex, shift;
	unsigned small = 32000, large = 0;

	/* Determine largest and smallest pattern sizes */
	for (i = 0; i < ps->msNumPatterns; i++) {
		if (ps->msPatArray[i].psLen < small)
			small = ps->msPatArray[i].psLen;
		if (ps->msPatArray[i].psLen > large)
			large = ps->msPatArray[i].psLen;
	}
	m = (strlen_t) small;
	if (m > 255)
		m = 255;
	ps->msShiftLen = m;

	/* Initialze the default shift table. Max shift of 256 characters */
	for (i = 0; i < 256; i++) {
		ps->msShift[i] = m;
	}

	/*  Multi-Pattern BAD CHARACTER SHIFT */
	for (i = 0; i < ps->msNumPatterns; i++) {
		for (k = 0; k < m; k++) {
			shift = (strlen_t) (m - 1 - k);
			if (shift > 255)
				shift = 255;
			cindex = ps->msPatArray[i].psPat[k];
			if (shift < ps->msShift[cindex])
				ps->msShift[cindex] = shift;
		}
	}
}

/*
** Prep and Build a Bad Word Shift table
*/
static void mbmPrepBadWordTable(struct mwm * ps)
{
	int i;
	strlen_t k, m, cindex;
	unsigned small = 32000, large = 0;
	unsigned shift;
	ps->msShift2 =
	    (unsigned char *) malloc(BWSHIFTABLESIZE * sizeof (char));
	if (!ps->msShift2)
		return;

	/* Determine largest and smallest pattern sizes */
	for (i = 0; i < ps->msNumPatterns; i++) {
		if (ps->msPatArray[i].psLen < small)
			small = ps->msPatArray[i].psLen;
		if (ps->msPatArray[i].psLen > large)
			large = ps->msPatArray[i].psLen;
	}
	m = (strlen_t) small;	/* Maximum Boyer-Moore Shift */

	/* Limit the maximum size of the smallest pattern to 255 bytes */
	if (m > 255)
		m = 255;
	ps->msShiftLen = m;

	/* Initialze the default shift table. */
	for (i = 0; i < BWSHIFTABLESIZE; i++) {
		ps->msShift2[i] = (unsigned) (m - 1);
	}
	/* Multi-Pattern Bad Word Shift Table Values */
	for (i = 0; i < ps->msNumPatterns; i++) {
		for (k = 0; k < m - 1; k++) {
			shift = (strlen_t) (m - 2 - k);
			if (shift > 255)
				shift = 255;
			cindex =
			    (ps->msPatArray[i].
			     psPat[k] | (ps->msPatArray[i].psPat[k + 1] << 8));
			if (shift < ps->msShift2[cindex])
				ps->msShift2[cindex] = shift;
		}
	}
}

/*
**  Print some Pattern Stats
*/
void mwmShowStats(void *pv)
{
	struct mwm *ps = (struct mwm *) pv;
	int i;
	printf("Pattern Stats\n");
	printf("-------------\n");
	printf("Patterns   : %d\n", ps->msNumPatterns);
	printf("Average    : %d chars\n", ps->msAvg);
	printf("Smallest   : %d chars\n", ps->msSmallest);
	printf("Largest    : %d chars\n", ps->msLargest);
	printf("Total chars: %d\n", ps->msTotal);
	for (i = 0; i < ps->msLargest + 1; i++) {
		if (ps->msLengths[i])
			printf("Len[%d] : %d patterns\n", i, ps->msLengths[i]);
	}
	printf("\n");
}

/*
**  Calc some pattern length stats
*/
static void mwmAnalyzePattens(struct mwm * ps)
{
	int i;
	ps->msLengths = (int *) malloc(sizeof (int) * (ps->msLargest + 1));
	if (ps->msLengths) {
		memset(ps->msLengths, 0, sizeof (int) * (ps->msLargest + 1));
		for (i = 0; i < ps->msNumPatterns; i++) {
			ps->msLengths[ps->msPatArray[i].psLen]++;
		}
	}
}

/*
**  Selects the bad Word Algorithm over the bad Character algorithm
**  This should be called before mwmPrepPatterns
*/
void mwmLargeShifts(void *pv, int flag)
{
	struct mwm *ps = (struct mwm *) pv;
	ps->msLargeShifts = flag;
}

/*
** mwmGetNpatterns::
*/
int mwmGetNumPatterns(void *pv)
{
	struct mwm *p = (struct mwm *) pv;
	return p->msNumPatterns;
}

/*
*  Assign the rule mask via an API
*/
void mwmSetRuleMask(void *pv, BITOP * rm)
{
	struct mwm *ps = (struct mwm *) pv;
	ps->RuleMask = rm;
}

/*
**  
**  No Bad Character Shifts
**  Handles pattern groups with one byte patterns
**  Uses 2 byte hash table to group patterns
**
*/
static int
mwmSearchExNoBC(struct mwm * ps, unsigned char *Tx, int n,
		int (*match) (unsigned id, int index, void *data), void *data)
{
	int Tleft, index, nfound, k;
	unsigned char *T, *Tend, *B;
	struct mwm_pat *patrn, *patrnEnd;
	nfound = 0;
	Tleft = n;
	Tend = Tx + n;

	/* Test if text is shorter than the shortest pattern */
	if ((unsigned) n < ps->msShiftLen)
		return 0;

	/*  Process each suffix of the Text, left to right, incrementing T so T = S[j] */
	for (T = Tx, B = Tx + ps->msShiftLen - 1; B < Tend; T++, B++, Tleft--) {

		/* Test for single char pattern matches */
		if ((index = ps->msHash1[*T]) != (hash_t) - 1) {
			patrn = &ps->msPatArray[index];
			patrnEnd = patrn + ps->msNumArray1[index];
			for (; patrn < patrnEnd; patrn++) {
				if (!patrn->psNoCase) {
					if (patrn->psPatCase[0] == case_normalise(*Tx)) {
						nfound++;
						if (match
						    (patrn->psID,
						     (int) (T - Tx), data))
							return nfound;
					}
				} else {
					nfound++;
					if (match
					    (patrn->psID, (int) (T - Tx), data))
						return nfound;
				}
			}
		}

		/* Test for last char in Text, one byte pattern test was done above, were done. */
		if (Tleft == 1)
			return nfound;

		/* Test if the 2 char prefix of this suffix shows up in the hash table */
		if ((index =
		     ps->msHash[((*T) << 8) | *(T + 1)]) == (hash_t) - 1)
			continue;

		/* Process the Hash Group Patterns against the current Text Suffix */
		patrn = &ps->msPatArray[index];
		patrnEnd = patrn + ps->msNumArray[index];

		/*  Match Loop - Test each pattern in the group against the Text */
		for (; patrn < patrnEnd; patrn++) {
			unsigned char *p, *q;

			/* Test if this Pattern is to big for Text, not a possible match */
			if ((unsigned) Tleft < patrn->psLen)
				continue;

			/* Test if this rule has been hit already - ignore it if it has */
			if (ps->RuleMask) {
				if (boIsBitSet(ps->RuleMask, patrn->psIID))
					continue;
			}

			/* Setup the reverse string compare */
			k = patrn->psLen - HASHBYTES16 - 1;
			q = patrn->psPat + HASHBYTES16;
			p = T + HASHBYTES16;

			/* Compare strings backward, unrolling does not help in perf tests. */
			while (k >= 0 && (q[k] == p[k]))
				k--;

			/* Test for a Match ! */
			if (k < 0) {
				if (!patrn->psNoCase) {
					if (!memcmpi
					    (patrn->psPatCase, T,
					     patrn->psLen)) {
						nfound++;
						if (match
						    (patrn->psID,
						     (int) (T - Tx), data))
							return nfound;
					}
				} else {
					nfound++;
					if (match
					    (patrn->psID, (int) (T - Tx), data))
						return nfound;
				}
			}
		}
	}
	return nfound;
}

/*
**
**  Uses Bad Character Shifts
**  Handles pattern groups with 2 or more bytes per pattern
**  Uses 2 byte hash table to group patterns
**
*/
static int
mwmSearchExBC(struct mwm * ps, unsigned char *Tx, int n,
	      int (*match) (unsigned id, int index, void *data), void *data)
{
	int Tleft, index, nfound, k, tshift;
	unsigned char *T, *Tend, *B;
	struct mwm_pat *patrn, *patrnEnd;
	nfound = 0;
	Tleft = n;
	Tend = Tx + n;

	/* Test if text is shorter than the shortest pattern */
	if ((unsigned) n < ps->msShiftLen)
		return 0;

	/*  Process each suffix of the Text, left to right, incrementing T so T = S[j] */
	for (T = Tx, B = Tx + ps->msShiftLen - 1; B < Tend; T++, B++, Tleft--) {

		/* Multi-Pattern Bad Character Shift */
		while ((tshift = ps->msShift[*B]) > 0) {
			B += tshift;
			T += tshift;
			Tleft -= tshift;
			if (B >= Tend)
				return nfound;
			tshift = ps->msShift[*B];
			B += tshift;
			T += tshift;
			Tleft -= tshift;
			if (B >= Tend)
				return nfound;
		}

		/* Test for last char in Text, one byte pattern test was done above, were done. */
		if (Tleft == 1)
			return nfound;

		/* Test if the 2 char prefix of this suffix shows up in the hash table */
		if ((index =
		     ps->msHash[((*T) << 8) | *(T + 1)]) == (hash_t) - 1)
			continue;

		/* Process the Hash Group Patterns against the current Text Suffix */
		patrn = &ps->msPatArray[index];
		patrnEnd = patrn + ps->msNumArray[index];

		/*  Match Loop - Test each pattern in the group against the Text */
		for (; patrn < patrnEnd; patrn++) {
			unsigned char *p, *q;

			/* Test if this Pattern is to big for Text, not a possible match */
			if ((unsigned) Tleft < patrn->psLen)
				continue;

			/* Test if this rule has been hit already - Ignore it if it has */
			if (ps->RuleMask) {
				if (boIsBitSet(ps->RuleMask, patrn->psIID))
					continue;
			}

			/* Setup the reverse string compare */
			k = patrn->psLen - HASHBYTES16 - 1;
			q = patrn->psPat + HASHBYTES16;
			p = T + HASHBYTES16;

			/* Compare strings backward, unrolling does not help. */
			while (k >= 0 && (q[k] == p[k]))
				k--;

			/* Test for a Match ! */
			if (k < 0) {
				if (!patrn->psNoCase) {
					if (!memcmpi
					    (patrn->psPatCase, T,
					     patrn->psLen)) {
						nfound++;
						if (match
						    (patrn->psID,
						     (int) (T - Tx), data))
							return nfound;
					}
				} else {
					nfound++;
					if (match
					    (patrn->psID, (int) (T - Tx), data))
						return nfound;
				}
			}
		}
	}
	return nfound;
}

/*
**
**  Uses Bad Word Shifts
**  Handles pattern groups with 2 or more bytes per pattern
**  Uses 2 byte hash table to group patterns
**
*/
static int
mwmSearchExBW(struct mwm * ps, unsigned char *Tx, int n,
	      int (*match) (unsigned id, int index, void *data), void *data)
{
	int Tleft, index, nfound, k, tshift;
	unsigned char *T, *Tend, *B;
	struct mwm_pat *patrn, *patrnEnd;
	nfound = 0;
	Tleft = n;
	Tend = Tx + n;

	/* Test if text is shorter than the shortest pattern */
	if ((unsigned) n < ps->msShiftLen)
		return 0;

	/*  Process each suffix of the Text, left to right, incrementing T so T = S[j] */
	for (T = Tx, B = Tx + ps->msShiftLen - 1; B < Tend; T++, B++, Tleft--) {

		/* Multi-Pattern Bad Word Shift */
		tshift = ps->msShift2[((*B) << 8) | *(B - 1)];
		while (tshift) {
			B += tshift;
			T += tshift;
			Tleft -= tshift;
			if (B >= Tend)
				return nfound;
			tshift = ps->msShift2[((*B) << 8) | *(B - 1)];
		}

		/* Test for last char in Text, we are done, one byte pattern test was done above. */
		if (Tleft == 1)
			return nfound;

		/* Test if the 2 char prefix of this suffix shows up in the hash table */
		if ((index =
		     ps->msHash[((*T) << 8) | *(T + 1)]) == (hash_t) - 1)
			continue;

		/* Process the Hash Group Patterns against the current Text Suffix */
		patrn = &ps->msPatArray[index];
		patrnEnd = patrn + ps->msNumArray[index];

		/*  Match Loop - Test each pattern in the group against the Text */
		for (; patrn < patrnEnd; patrn++) {
			unsigned char *p, *q;

			/* Test if this Pattern is to big for Text, not a possible match */
			if ((unsigned) Tleft < patrn->psLen)
				continue;

			/* Test if this rule has been hit already - Ignore it if it has */
			if (ps->RuleMask) {
				if (boIsBitSet(ps->RuleMask, patrn->psIID))
					continue;
			}

			/* Setup the reverse string compare */
			k = patrn->psLen - HASHBYTES16 - 1;
			q = patrn->psPat + HASHBYTES16;
			p = T + HASHBYTES16;

			/* Compare strings backward, unrolling does not help. */
			while (k >= 0 && (q[k] == p[k]))
				k--;

			/* Test for a Match ! */
			if (k < 0) {
				if (!patrn->psNoCase) {
					if (!memcmpi
					    (patrn->psPatCase, T,
					     patrn->psLen)) {
						nfound++;
						if (match
						    (patrn->psID,
						     (int) (T - Tx), data))
							return nfound;
					}
				} else {
					nfound++;
					if (match
					    (patrn->psID, (int) (T - Tx), data))
						return nfound;
				}
			}
		}
	}
	return nfound;
}

/* Display function for testing */
static void show_bytes(unsigned n, unsigned char *p)
{
	int i;
	for (i = 0; i < (int) n; i++) {
		if (p[i] >= 32 && p[i] <= 127)
			printf("%c", p[i]);

		else
			printf("\\x%2.2X", p[i]);
	}
}

/*
**   Display patterns in this group
*/
void mwmGroupDetails(void *pv)
{
	struct mwm *ps = (struct mwm *) pv;
	int index, i, m, gmax = 0, total = 0, gavg = 0, subgroups;
	static int k = 0;
	struct mwm_pat *patrn, *patrnEnd;
	printf("*** MWM-Pattern-Group: %d\n", k++);
	subgroups = 0;
	for (i = 0; i < 65536; i++) {
		if ((index = ps->msHash[i]) == (hash_t) - 1)
			continue;
		patrn = &ps->msPatArray[index];	/* 1st pattern of hash group is here */
		patrnEnd = patrn + ps->msNumArray[index];	/* never go here... */
		printf("  Sub-Pattern-Group: %d-%d\n", subgroups, i);
		subgroups++;
		for (m = 0; patrn < patrnEnd; m++, patrn++) {	/* Test all patterns in the group */
			printf("   Pattern[%d] : ", m);
			show_bytes(patrn->psLen, patrn->psPat);
			printf("\n");
		}
		if (m > gmax)
			gmax = m;
		total += m;
		gavg = total / subgroups;
	}
	printf("Total Group Patterns    : %d\n", total);
	printf("  Number of Sub-Groups  : %d\n", subgroups);
	printf("  Sub-Group Max Patterns: %d\n", gmax);
	printf("  Sub-Group Avg Patterns: %d\n", gavg);
}

/*
**
** mwmPrepPatterns::    Prepare the pattern group for searching
**
*/
int mwmPrepPatterns(void *pv)
{
	struct mwm *ps = (struct mwm *) pv;
	int kk;
	struct mwm_pat *plist;

	/* Build an array of pointers to the list of Pattern nodes */
	ps->msPatArray =
	    (struct mwm_pat *) calloc(sizeof (struct mwm_pat),
					  ps->msNumPatterns);
	if (!ps->msPatArray) {
		return -1;
	}
	ps->msNumArray =
	    (strlen_t *) calloc(sizeof (strlen_t), ps->msNumPatterns);
	if (!ps->msNumArray) {
		return -1;
	}

	/* Copy the list node info into the Array */
	for (kk = 0, plist = ps->plist;
	     plist != NULL && kk < ps->msNumPatterns; plist = plist->next) {
		memcpy(&ps->msPatArray[kk++], plist,
		       sizeof (struct mwm_pat));
	}
	mwmAnalyzePattens(ps);

	/* Sort the patterns */
	qsort(ps->msPatArray, ps->msNumPatterns,
	      sizeof (struct mwm_pat), sortcmp);

	/* Build the Hash table, and pattern groups, per Wu & Manber */
	mwmPrepHashedPatternGroups(ps);

	/* Select the Pattern Matcher Class */
	if (ps->msNumPatterns < 5) {
		ps->msMethod = MTH_BM;
	}

	else {
		ps->msMethod = MTH_MWM;
	}

	/* Setup Wu-Manber */
	if (ps->msMethod == MTH_MWM) {

		/* Build the Bad Char Shift Table per Wu & Manber */
		mwmPrepBadCharTable(ps);

		/* Build the Bad Word Shift Table per Wu & Manber */
		if ((ps->msShiftLen > 1) && ps->msLargeShifts) {
			mbmPrepBadWordTable(ps);
		}

		/* Min patterns is 1 byte */
		if (ps->msShiftLen == 1) {
			ps->search = mwmSearchExNoBC;
		}

		/* Min patterns is >1 byte */
		else if ((ps->msShiftLen > 1) && !ps->msLargeShifts) {
			ps->search = mwmSearchExBC;
		}

		/* Min patterns is >1 byte - and we've been asked to use a 2 byte bad words shift instead. */
		else if ((ps->msShiftLen > 1) && ps->msLargeShifts
			 && ps->msShift2) {
			ps->search = mwmSearchExBW;
		}

		/* Min patterns is >1 byte */
		else {
			ps->search = mwmSearchExBC;
		}
	}

	/* Initialize the Boyer-Moore Pattern data */
	if (ps->msMethod == MTH_BM) {
		int i;

		/* Allocate and initialize the BMH data for each pattern */
		for (i = 0; i < ps->msNumPatterns; i++) {
			ps->msPatArray[i].psBmh =
			    hbm_prep(ps->msPatArray[i].psPat,
				     ps->msPatArray[i].psLen);
		}
	}
	return 0;
}

/*
** Search a body of text or data for paterns 
*/
int mwmSearch(void *pv, unsigned char *T, int n,
	  int (*match) (unsigned id, int index, void *data), void *data)
{
	struct mwm *ps = (struct mwm *) pv;

	if (ps->msMethod == MTH_BM) {
		/* Boyer-Moore  */
		int i, nfound = 0;
		unsigned char *Tx;
		for (i = 0; i < ps->msNumPatterns; i++) {
			Tx = hbm_match(ps->msPatArray[i].psBmh, T, n);
			if (Tx) {

				/* If we are case sensitive, do a final exact match test */
				if (!ps->msPatArray[i].psNoCase) {
					if (memcmp
					    (ps->msPatArray[i].psPatCase,
					     &T[Tx - T],
					     ps->msPatArray[i].psLen))
						continue;	/* no match, next pattern please */
				}
				nfound++;
				if (match
				    (ps->msPatArray[i].psID,
				     (int) (Tx - T), data))
					return nfound;
			}
		}
		return nfound;
	} else {
		/* Wu-Manber */
		return ps->search(ps, T, n, match, data);
	}
}
