/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
*  o capture_cleanup() - stop the packet capture
*  o capture_init() - register cleanup
*  o capture_stop() - finish the current capture
*  o capture_interrupt() - interrupt the current capture
*  o capture_setup() - initialise the capture
*  o capture_timestamp() - get a (cpu) timestamp
*  o capture_go() - packet capture main loop
*/
#include <config.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif
#include <unistd.h>

#include <firestorm.h>
#include <cleanup.h>
#include <packet.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <capture.h>

struct capture capture;

/* Make sure to call the capdev cleanup function before we quit */
void capture_cleanup(int state, void *priv)
{
	if ( capture.priv )
		capture.capdev->end(capture.priv);
	if ( capture.args )
		free(capture.args);
}

void __attribute__((constructor)) capture_init(void) 
{
	cleanup_add(capture_cleanup, NULL);
}

/* Called by the interrupt handlers for clean shutdown */
void capture_stop(void)
{
	capture.state=CAP_STATE_STOP;
}

void capture_interrupt(void)
{
	capture.state=CAP_STATE_INT;
}

/* Setup the capture structures to get going */
void capture_setup(char *type, char *args)
{
	struct capdev *cd;
	char *a;

	if ( !type || !args ) {
		cleanup(EXIT_ERR, "capture: wtf?!");
	}

	if ( !(cd=capdev_find(type)) ) {
		cleanup(EXIT_ERR, "capture: cant find '%s'", type);
	}

	if ( !(a=strdup(args)) ) {
		cperror("strdup");
	}

	capture.capdev=cd;
	capture.args=a;
	if ( !(capture.priv=capture.capdev->init(args)) )
		cleanup(EXIT_ERR, "capture: failed to initialise");
	capture.state=CAP_STATE_INIT;
}

/* Get a timestamp */
static inline void capture_timestamp(struct timeval *tv)
{
#ifdef HAVE_GETRUSAGE
	static struct rusage r;
	getrusage(RUSAGE_SELF, &r);
	tv->tv_sec=r.ru_utime.tv_sec;
	tv->tv_usec=r.ru_utime.tv_usec;
#else
	gettimeofday(tv, NULL);
#endif
}

/* Go in to the packet capture main loop */
void capture_go(void)
{
	struct timeval tv, tv2;
	struct timeval end;

	if ( capture.state!=CAP_STATE_INIT ) {
		cleanup(EXIT_ERR, "No captures specified");
	}

	/* Ready to go */
	mesg(M_INFO,"capture: %s[%s]: started",
		capture.capdev->name, capture.args);

	/* Do the capture */
	capture.state=CAP_STATE_CAPTURE;
	capture_timestamp(&tv);
	do {
		capture.capdev->go(capture.priv, &capture);
		if ( capture.state==CAP_STATE_INT ) {
			/* An interrupt was recieved */
			capture.state=CAP_STATE_CAPTURE;
			alert_hup();
		}
	} while ( capture.state==CAP_STATE_CAPTURE );
	capture_timestamp(&tv2);
	capture.state=CAP_STATE_ENDED;

	/* All done */
	mesg(M_INFO,"capture: %s[%s]: stopped: %llu packets",
		capture.capdev->name,capture.args,the_serial);

	/* Calculate elapsed times */
	end.tv_sec=(tv2.tv_sec-tv.tv_sec)-1;
	end.tv_usec=(1000000+tv2.tv_usec)-tv.tv_usec;
	end.tv_sec+=end.tv_usec/1000000;
	end.tv_usec%=1000000;

	/* TODO: Get CPU time as well as wall time */
	mesg(M_DEBUG,"ELAPSED "
#ifdef HAVE_GETRUSAGE
		"CPU"
#endif
		": %lu.%.6lu", end.tv_sec, end.tv_usec);
}
