/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* TODO:
*  o Allow for specifying output file for elog concatenation
*  o Check for conflicting arguments
*/
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <stdarg.h>
#include <unistd.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#include <fcntl.h>
#include <string.h>
#include <netinet/in.h>

#include <firestorm.h>
#include <packet.h>
#include <cleanup.h>
#include <loader.h>
#include <alert.h>
#include <args.h>
#include <target.h>
#include <signature.h>
#include <decode.h>
#include <matcher.h>
#include <detect.h>
#include <capture.h>
#include <elog.h>
#include <elog_read.h>

/* Indexing library */
#include <mpool.h>
#include <fbuf.h>
#include <idx.h>
#include <i_idx.h>
#include <i_ordered.h>
#include <q_idx.h>
#include <q_filter.h>

#include <firecat.h>

#ifdef HAVE_GETOPT_LONG
static const struct option opts[]={
	{"plugin-dir", 	1, 0, 'p'},
	{"options", 	1, 0, 'o'},
	{"format", 	1, 0, 'f'},
	{"index", 	0, 0, 'i'},
	{"list",	0, 0, 'l'},
	{"query", 	1, 0, 'q'},
	{"help", 	0, 0, 'h'},
	{NULL, 0, 0, 0}
};
#endif

/* Store argv[0] */
static char *cmd;

static int verbose_flag;

/* for MODE_CAT */
static struct target *t;
static void *t_priv;
static char *oformat;
static char *p_options;
static size_t cur_ofs;

/* for MODE_IDX */
static char *query;
static struct q_filter *filter;

/* Functions we may need to fake */
int return_one(void) {return 1;}
void return_void(void) {return;}
void *return_null(void) {return NULL;}
void serial_number(serial_t *s){*s=0;}

struct {
	const char *name;
	void *data;
}exports[]={
	/* Generic */
	{"mesg",		mesg},
	{"args.parse", 		args_parse},
	{"serial_number",	serial_number},

	/* Object lookup */
	{"decode.subproto", 	decode_subproto},
	{"decode.proto", 	decode_proto},
	{"matcher.find",	return_null},

	/* Needed for signature engines */
	{"alert",		return_void},
	{"generator.add",	return_one},
	{"detect.add_sig",	return_one},
	{"detect.optimize",	return_void},
	{"detect.free_sig",	return_void},
	{"detect",		return_void},

	{NULL,NULL}
};

/* Only print wanings,errors,and critical messages */
void mesg(unsigned char code, char *fmt, ...)
{
	va_list va;

	if ( !verbose_flag && code > M_WARN )
		return;

	/* Use stderr cos output may use stdout */
	va_start(va, fmt);
	fprintf(stderr, "%s: ", cmd);
	vfprintf(stderr, fmt, va);
	fprintf(stderr, "\n");
	va_end(va);
}

void *plugin_callback(const char *item)
{
	u_int32_t i;

	for(i=0; exports[i].name; i++) {
		if ( !strcmp(exports[i].name, item) ) {
			return exports[i].data;
		}
	}

	fprintf(stderr, "resource: request for unknown object: '%s'\n", item);
	return NULL;
}

void get_plugins(void)
{
	/* Load target plugins */
#if 1
	loader_load_dir(PLUGIN_DIR "/protocols");
	loader_load_dir(PLUGIN_DIR "/output");
#else
	loader_load_dir("./decode_plugins/.libs");
	loader_load_dir("./target_plugins/.libs");
#endif

	decode_load();
	target_load();
}

/* Display usage message and quit */
void usage(int ecode)
{
	extern struct target *targets;
	struct target *t;

	fprintf(stderr, "Usage: %s [-f format] [-o \"options\"] [FILE]...\n", cmd);
	fprintf(stderr, "Concatenates and converts firestorm extended log (elog) files.\n\n");
	fprintf(stderr, "    --format, -f      Choose a coversion format\n");
	fprintf(stderr, "    --options, -o     Plugin specific option string\n");
	fprintf(stderr, "    --plugin-dir, -p  Load extra plugins from this directory\n");
	fprintf(stderr, "    --index, -i       Create query indexes\n");
	fprintf(stderr, "    --list, -l        View index information for a file\n");
	fprintf(stderr, "    --query, -q       Select only certain packets\n");
	fprintf(stderr, "    --verbose, -v     Display debugging messages to stderr\n");
	fprintf(stderr, "    --help, -h        This message\n\n");
	fprintf(stderr, "With no FILE, or when FILE is -, read standard input.\n\n");
	fprintf(stderr, "Loading plugins from: %s\n", PLUGIN_DIR "/protocols");
	fprintf(stderr, "Loading plugins from: %s\n\n", PLUGIN_DIR "/output");
	fprintf(stderr, "Available conversion formats:\n");
	fprintf(stderr, "    elog - no conversion\n");

	get_plugins();

	/* Display available target plugins */
	for(t=targets; t; t=t->next) {
		fprintf(stderr, "    %s - %s\n", t->name, t->desc);
	}

	fprintf(stderr, "\nReport bugs to <gianni@scaramanga.co.uk>\n");
	cleanup(ecode,NULL);
}

int cbfn_target(struct elog_pkt *p, void *data)
{
	cur_ofs = p->ofs;
	return elog_pkt_alert(p, t->target, data);
}

static struct fbuf eobuf;

/* TODO: allow file=foo */
static void init_elog(char *args)
{
	struct efile_hdr fhdr;
	char *fn = "(stdout)";
	
	if ( fbuf_attach(&eobuf, STDOUT_FILENO, 32768) < 0 )
		cleanup(EXIT_ERR, "fbuf_attach(): %s", get_err());

	fhdr.magic = htonl(EF_MAGIC);
	fhdr.flags = 0;
	fhdr.vers_minor = EF_VERS_MIN;
	fhdr.vers_major = EF_VERS_MAJ;
	if ( fbuf_write(&eobuf, &fhdr, sizeof(fhdr)) < 0 )
		cleanup(EXIT_ERR, "fbuf_write(): %s",
			errno ? get_err() : "hdr truncated");
}

static void fini_elog(void)
{
	if ( fbuf_close(&eobuf) < 0 )
		cleanup(EXIT_ERR, "fbuf_close(): %s", get_err());
}

static int cbfn_elog(struct elog_pkt *p, void *data)
{
	char *pkt = p->file->map + p->ofs;
	size_t len = ntohl(p->hdr->h.reclen);

	if ( fbuf_write(&eobuf, pkt, len) < 0 )
		cleanup(EXIT_ERR, "fbuf_write(): %s", get_err());
	return 1;
}

/* Process a file with a target plugin */
void cat_file(char *fn, int fd)
{
	void *ep;

	if ( !(ep=elog_new(fd)) ) {
		cleanup(EXIT_ERR, "%s: elog_new(): %s", fn, get_err());
		return;
	}

	if ( t == NULL ) {
		elog_set_alert_fn(ep, cbfn_elog);
	}else{
		elog_set_alert_fn(ep, cbfn_target);
	}

	if ( elog_run(ep, t_priv)<=0 ) {
		elog_free(ep);
		cleanup(EXIT_ERR, NULL);
	}

	elog_free(ep);
}

/* Basically we just load and initialise the required target plugin */
void cat_init(void)
{
	get_plugins();

	/* Use elog if none specified */
	if ( !oformat ) {
		oformat="elog";
		init_elog(p_options);
		return;
	}

	/* Get the right output plugin */
	if ( !(t=target_find(oformat)) ) {
		mesg(M_ERR, "%s: Cannot find specified output format",
			oformat);
		cleanup(EXIT_ERR,NULL);
	}

	/* Initialise it with user specified arguments*/
	if ( !t->validate(p_options, &t_priv) ) {
		mesg(M_ERR, "%s: failed to initialize output plugin",
			oformat);
		cleanup(EXIT_ERR,NULL);
	}
}

void cat_fini(void)
{
	if ( t == NULL ) {
		fini_elog();
		return;
	}

	if ( !t->cleanup(t_priv) )
		cleanup(EXIT_ERR,NULL);
}

int idx_target_fn(struct alert *a, struct packet *p, void *priv)
{
	struct i_ordered *dbs = t_priv;
	struct proto_field *f;
	idx_key_t v;
	int i;

	/* Do non-layer stuff */
	for(f=builtin_fields; f->name; f++) {
		if ( !f->get_key.num(p, a, -1, &v) )
			goto err;

		if ( i_ordered_insert(f->ornament, v, cur_ofs) < 0 )
			goto err;
	}

	/* Do the per-layer stuff */
	for(i=0; i < p->llen; i++) {
		if ( p->layer[i].proto == NULL )
			continue;

		if ( p->layer[i].proto->fields == NULL )
			continue;

		for(f=p->layer[i].proto->fields; f->name; f++) {
			if ( !f->get_key.num(p, a, i, &v) )
				goto err;

			if ( i_ordered_insert(f->ornament, v, cur_ofs) < 0 )
				goto err;
		}
	}

	return 1;

err:
	cleanup(EXIT_ERR, "Error adding %s field to index", f->name);
}

void lst_file(char *fn, int fd)
{
	size_t fnlen = strlen(fn);
	char fnbuf[fnlen + 5];
	struct q_db *db;
	struct stat st1, st2;
	char *efn=NULL;
	int efd;
	int ood = 0;
	int i;

	/* See if we have an index only, or an elog plus an index */
	if ( fnlen > 4 && strncmp(fn + fnlen - 4, ".idx", 4) ) {
		snprintf(fnbuf, sizeof(fnbuf), "%s.idx", fn);
		efn = fn;
		efd = fd;
		fn = fnbuf;

		if ( fstat(fd, &st1) )
			cleanup(EXIT_ERR, "%s: fsync(): %s", efn, get_err());

		fd = open(fn, O_RDONLY);
		if ( fd < 0 )
			cleanup(EXIT_ERR, "%s: open(): %s", fn, get_err());

		if ( fstat(fd, &st2) )
			cleanup(EXIT_ERR, "%s: fsync(): %s", fn, get_err());

		if ( st1.st_mtime >= st2.st_mtime )
			ood = 1;
	}

	/* Print out filenames */
	db = q_idx_open(fd);
	if ( db == NULL )
		cleanup(EXIT_ERR, "%s: q_idx_open(): %s", fn, get_err());

	printf("idx: %s\n", fn);
	if ( efn ) {
		printf("elog: %s\n", efn);
		printf("status: %s\n", ood ? "out of date" : "up to date");
	}

	/* Print all of the fields */
	for(i=0; i < db->num_idx; i++) {
		const char *type;

		switch(db->idx[i].type) {
		case IDX_TYPE_ORDERED:
			type = "ordered";
			break;
		case IDX_TYPE_HASHED:
			type = "hashed";
			break;
		default:
			type = "invalid";
			break;
		}

		printf("field: name=%s num_keys=%u num_vals=%u type=%s\n",
				db->idx[i].name,
				db->idx[i].num_keys,
				db->idx[i].num_vals,
				type);
	}

	printf("\n");

	/* Close the files */
	close(fd);
	if ( efn )
		close(efd);
}

/* =====[ Index File Creation Interface ]===== */
struct target idx_target = {
	.name = "index",
	.target = idx_target_fn,
};

void idx_init(void)
{
	struct i_ordered *dbs;

	get_plugins();

	/* Setup the db structures */
	dbs = malloc(num_fields * sizeof(*dbs));
	if ( dbs == NULL )
		cperror("malloc");

	t = &idx_target;
	t_priv = dbs;
}

void idx_fini(void)
{
	if ( t_priv )
		free(t_priv);
}

void idx_file(char *fn, int fd)
{
	struct i_ordered *dbs = t_priv;
	struct proto_field *f;
	struct i_idx *idx;
	void *ep;
	char nbuf[strlen(fn) + 5];
	int ofd;

	/* Open the output file */
	snprintf(nbuf, sizeof(nbuf), "%s.idx", fn);
	ofd = open(nbuf, O_WRONLY|O_TRUNC|O_CREAT, 0640);
	if ( ofd < 0 ) {
		mesg(M_ERR, "%s: open(): %s", nbuf, get_err());
		goto err;
	}

	/* Open the input file */
	if ( !(ep=elog_new(fd)) ) {
		mesg(M_ERR, "%s: elog_new(): %s", fn, get_err());
		goto err;
	}

	/* Create the databases */
	for(f=fields; f; f=f->next, dbs++) {
		if ( i_ordered_new(dbs) < 0 )
			goto err;

		f->ornament = dbs;
	}

	/* Run through the file */
	elog_set_alert_fn(ep, cbfn_target);
	if ( elog_run(ep, t_priv)<=0 )
		goto err;

	elog_free(ep);

	/* Write the index file out to disk */
	idx = i_idx_new(num_fields);
	if ( idx == NULL )
		goto err;

	/* Add each field to the DB */
	for(f=fields, dbs=t_priv; f; f=f->next, dbs++) {
		if ( !i_idx_ordered_field(idx, f->name, dbs) )
			goto err;
	}

	/* Write the index file out, this syncs and closes the fd */
	i_idx_write_fd(idx, ofd);

	/* Free everything up */
	i_idx_free(idx);
	return;

err:
	cleanup(EXIT_ERR, "%s: error: %s", fn,
		errno ? get_err() : "unknown");
	return;
}

/* =====[ ELOG Query Interface ]===== */

void qry_init(void)
{
	filter = q_filter_new(0);
	if ( filter == NULL )
		cleanup(EXIT_ERR, "q_filter_new: %s", get_err());

	if ( !q_filter_query(filter, query) )
		cleanup(EXIT_ERR, "q_filter_query: %s", get_err());

	cat_init();
}

void qry_file(char *fn, int fd)
{
	size_t fnlen = strlen(fn);
	char fnbuf[fnlen + 5];
	struct stat st1, st2;
	void *ep;
	int ifd;
	int i;

	if ( fstat(fd, &st1) < 0 )
		cleanup(EXIT_ERR, "%s: fsync(): %s", fn, get_err());

	/* Open the elog file */
	ep = elog_new(fd);
	if ( ep == NULL )
		cleanup(EXIT_ERR, "%s: elog_new(): %s", fn, get_err());

	if ( t == NULL ) {
		elog_set_alert_fn(ep, cbfn_elog);
	}else{
		elog_set_alert_fn(ep, cbfn_target);
	}
	
	/* Open the index file */
	snprintf(fnbuf, sizeof(fnbuf), "%s.idx", fn);
	ifd = open(fnbuf, O_RDONLY);
	if ( ifd < 0 )
		cleanup(EXIT_ERR, "%s: open(): %s", fnbuf, get_err());

	if ( fstat(ifd, &st2) < 0 )
		cleanup(EXIT_ERR, "%s: fstat(): %s", fnbuf, get_err());

	/* Check the index is older than the elog */
	if ( st2.st_mtime < st1.st_mtime )
		cleanup(EXIT_ERR, "%s: %s is out of date", fn, fnbuf);

	/* Actually run the query */
	if ( !q_filter_run(filter, ifd) )
		cleanup(EXIT_ERR, "q_filter_run() failed");

	/* Process the results */
	for(i=0; i < filter->num_recs; i++) {
		if ( elog_pkt(ep, filter->result[i], t_priv) <= 0 )
			cleanup(EXIT_ERR, NULL);
	}

	q_filter_free_result(filter);

	elog_free(ep);
	close(ifd);
}

/* Define the various modes of operation */
static struct firecat_mode mode[]={
	[ MODE_CAT ] { .init=cat_init, .file=cat_file, .fini=cat_fini },
	[ MODE_IDX ] { .init=idx_init, .file=idx_file, .fini=idx_fini },
	[ MODE_QER ] { .init=qry_init, .file=qry_file, .fini=cat_fini },
	[ MODE_LST ] { .init=NULL, .file=lst_file, .fini=NULL },
};

int main(int argc, char **argv)
{
	int m=MODE_CAT;
	int c, opti;
	int fd;
	int i;

	/* Initialise what we need to */
	cmd=argv[0];
	loader_init(plugin_callback);

	/* Do command line args */
#ifdef HAVE_GETOPT_LONG
	while( (c=getopt_long(argc, argv, "p:q:if:lo:vh?",
		opts, &opti))!=-1 )
#else
	while( (c=getopt(argc, argv, "p:q:if:lo:vh?"))!=-1 )
#endif
	{
		switch (c) {
		case 'i':
			m = MODE_IDX;
			break;
		case 'l':
			m = MODE_LST;
			break;
		case 'q':
			m = MODE_QER;
			query = optarg;
			break;
		case 'p':
			loader_load_dir(optarg);
			break;
		case 'f':
			if ( oformat )
				usage(EXIT_ERR);
			oformat=optarg;
			break;
		case 'o':
			if ( p_options )
				usage(EXIT_ERR);
			p_options=optarg;
			break;
		case 'v':
			verbose_flag = 1;
			break;
		case 'h':
		case '?':
			usage(EXIT_OK);
			break;
		default:
			usage(EXIT_ERR);
		};
	}

	/* TODO: Check for conflicting parameters */

	if ( mode[m].init )
		mode[m].init();

	/* cat semantics. if no file specified, or file is -, read
	 * standard input */
	if ( optind >= argc ) {
		mode[m].file("(stdin)", STDIN_FILENO);
	}

	for(i=optind; i<argc; i++ ) {
		/* use stdin for - */
		if ( !strcmp(argv[i],"-") ) {
			mode[m].file("(stdin)", STDIN_FILENO);
		}else{
			/* Open the input file */
			if ( (fd=open(argv[i],O_RDONLY))<0 ) {
				mesg(M_ERR, "%s: open(): %s",
					argv[i], get_err());
				cleanup(EXIT_ERR,NULL);
			}

			mode[m].file(argv[i], fd);
			close(fd);
		}
	}

	if ( mode[m].fini )
		mode[m].fini();

	cleanup(EXIT_OK,NULL);
	return 0;
}
