/*
* This file is part of Firestorm NIDS
* Copyright (c) 2003 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This is a GtkTreeModel that maps an elog file and allows it to be
* displayed efficiently inside of a GtkTreeView.
*
* TODO:
*  o Combine with query API for filtering/sorting
*  o Implement drag and drop interfaces?
*  o GAL like display of time
*
* Additional columns
*  o Source port
*  o Dest port
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <time.h>
#include <gnome.h>
#include <errno.h>

#include <firestorm.h>
#include <args.h>
#include <packet.h>
#include <alert.h>
#include <signature.h>
#include <target.h>
#include <decode.h>
#include <elog.h>
#include <elog_read.h>
#include <g_etree_model.h>

static GdkPixbuf *icon;
static GObjectClass *parent_class;

static int col_cb_prio(struct alert *a, struct packet *p, GValue *v)
{
	int n=a->priority;
	char *prio[]={
		"unknown", "critical", "high", "medium", "low", "auto"
	};

	if ( n > 5 )
		n=0;

	g_value_set_static_string(v, prio[n]);
	return 1;
}

static int col_cb_src(struct alert *a, struct packet *p, GValue *v)
{
	struct in_addr ip;
	int i;

	for(i=0; i<p->llen; i++) {
		if ( p->layer[i].proto == NULL )
			continue;

		if ( !strcmp("ip", p->layer[i].proto->name) )
			break;
	}

	if ( i>=p->llen ) {
		g_value_set_static_string(v, "(no-ip)");
		return 1;
	}

	ip.s_addr = p->layer[i].h.ip->saddr;

	g_value_set_string(v, (const gchar *)inet_ntoa(ip));
	return 1;
}

static int col_cb_dst(struct alert *a, struct packet *p, GValue *v)
{
	struct in_addr ip;
	int i;

	for(i=0; i<p->llen; i++) {
		if ( p->layer[i].proto == NULL )
			continue;

		if ( !strcmp("ip", p->layer[i].proto->name) )
			break;
	}

	if ( i>=p->llen ) {
		g_value_set_static_string(v, "(no-ip)");
		return 1;
	}

	ip.s_addr = p->layer[i].h.ip->daddr;

	g_value_set_string(v, (const gchar *)inet_ntoa(ip));
	return 1;
}

static int col_cb_gen(struct alert *a, struct packet *p, GValue *v)
{
	g_value_set_static_string(v, a->gen->name);
	return 1;
}

static int col_cb_msg(struct alert *a, struct packet *p, GValue *v)
{
	g_value_set_static_string(v, a->alert);
	return 1;
}

static int col_cb_time(struct alert *a, struct packet *p, GValue *v)
{
	char buf[64];
	struct tm *rtime;

	rtime=localtime(&p->time.tv_sec);
	strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", rtime);
	g_value_set_string(v, (const gchar *)buf);

	return 1;
}

static struct {
	GType type;
	proc_target fn;
}cols[]={
	[ ETREE_ICON ] {G_TYPE_OBJECT, NULL}, /* special case */
	[ ETREE_PRIO ] {G_TYPE_STRING, (proc_target)col_cb_prio},
	[ ETREE_SRC  ] {G_TYPE_STRING, (proc_target)col_cb_src},
	[ ETREE_DST  ] {G_TYPE_STRING, (proc_target)col_cb_dst},
	[ ETREE_GEN  ] {G_TYPE_STRING, (proc_target)col_cb_gen},
	[ ETREE_MSG  ] {G_TYPE_STRING, (proc_target)col_cb_msg},
	[ ETREE_TIME ] {G_TYPE_STRING, (proc_target)col_cb_time},
};

/* dispatch to the current column callback */
static gint cur_column;
int column_cb(struct elog_pkt *pkt, void *user_data)
{
	return elog_pkt_alert(pkt, cols[cur_column].fn, user_data);
}

/* Rather silly utility function to load and scale a pixmap */
static GdkPixbuf *load_pixmap(char *fn)
{
	GdkPixbuf *t, *ret=NULL;

	if ( !(t=gdk_pixbuf_new_from_file(fn, NULL)) )
		return NULL;

	if ( !(ret=gdk_pixbuf_scale_simple(t, 24, 24,
		GDK_INTERP_HYPER)) ) {
		g_object_unref(t);
		return NULL;
	}

	g_object_unref(t);
	return ret;
}

static void etree_model_init(EtreeModel *etree)
{
	etree->stamp = g_random_int();
	etree->num_pkts = 0;
	etree->efile = NULL;
	etree->appbar = NULL;
}

static void etree_model_finalize(GObject *object)
{
	(*parent_class->finalize)(object);
}

static void etree_model_class_init(EtreeModelClass *class)
{
	GObjectClass *object_class;

	parent_class = g_type_class_peek_parent(class);
	object_class = (GObjectClass *)class;

	object_class->finalize = etree_model_finalize;
}

static GtkTreeModelFlags get_flags(GtkTreeModel *tree_model)
{
	g_return_val_if_fail(IS_ETREE_MODEL(tree_model), 0);
	return GTK_TREE_MODEL_ITERS_PERSIST | GTK_TREE_MODEL_LIST_ONLY;
}

static gint get_n_columns(GtkTreeModel *tree_model)
{
	g_return_val_if_fail(IS_ETREE_MODEL(tree_model), 0);
	return ETREE_MAX;
}

static GType get_column_type(GtkTreeModel *tree_model, gint idx)
{
	g_return_val_if_fail(IS_ETREE_MODEL(tree_model), 0);
	g_return_val_if_fail(idx < ETREE_MAX, 0);

	return cols[idx].type;
}

static gboolean get_iter(GtkTreeModel *tree_model,
			GtkTreeIter *iter,
			GtkTreePath *path)
{
	EtreeModel *e = (EtreeModel *)tree_model;
	gint i;

	g_return_val_if_fail(IS_ETREE_MODEL(tree_model), FALSE);

	i = gtk_tree_path_get_indices(path)[0];
	if ( i >= e->num_pkts )
		return FALSE;
	
	iter->stamp = e->stamp;
	iter->user_data = (gpointer)i;
	return TRUE;
}

static GtkTreePath *get_path(GtkTreeModel *tree_model,
				GtkTreeIter *iter)
{
	EtreeModel *e = (EtreeModel *)tree_model;
	GtkTreePath *ret;

	g_return_val_if_fail(IS_ETREE_MODEL(tree_model), NULL);
	g_return_val_if_fail((size_t)iter->user_data < e->num_pkts, NULL);

	ret = gtk_tree_path_new();
	gtk_tree_path_append_index(ret, (size_t)iter->user_data);
	return ret;
}

static void get_value(GtkTreeModel *tree_model,
			GtkTreeIter *iter,
			gint column,
			GValue *value)
{
	EtreeModel *e = (EtreeModel *)tree_model;
	size_t ofs;

	g_return_if_fail(IS_ETREE_MODEL(tree_model));
	g_return_if_fail(iter->stamp == e->stamp);
	g_return_if_fail((size_t)iter->user_data < e->num_pkts);
	g_return_if_fail(column < ETREE_MAX);

	g_value_init(value, cols[column].type);

	/* Special case for the icon, we don't need to
	 * read the packet to get the icon (just now) ;)
	 */
	if ( column == ETREE_ICON ) {
		g_value_set_object(value, icon);
		return;
	}

	/* Pass to the relevent callback */
	ofs = e->ofs[(size_t)iter->user_data];
	cur_column = column;
	elog_set_alert_fn(e->efile, column_cb);
	elog_pkt(e->efile, ofs, value);
}

static gboolean iter_next(GtkTreeModel *tree_model,
				GtkTreeIter *iter)
{
	EtreeModel *e = (EtreeModel *)tree_model;

	g_return_val_if_fail(IS_ETREE_MODEL(tree_model), FALSE);

	if ( (size_t)iter->user_data >= e->num_pkts-1 )
		return FALSE;

	iter->user_data = (gpointer)(((size_t)iter->user_data) + 1);
	return TRUE;
}

static gboolean iter_children(GtkTreeModel *tree_model,
				GtkTreeIter *iter,
				GtkTreeIter *parent)
{
	EtreeModel *e = (EtreeModel *)tree_model;

	g_return_val_if_fail(IS_ETREE_MODEL(tree_model), FALSE);

	if ( parent )
		return FALSE;

	if ( e->num_pkts ) {
		iter->stamp = e->stamp;
		iter->user_data = (size_t)0;
		return TRUE;
	}

	return FALSE;
}

static gboolean iter_has_child(GtkTreeModel *tree_model,
				GtkTreeIter *iter)
{
	return FALSE;
}

static gint iter_n_children(GtkTreeModel *tree_model,
				GtkTreeIter *iter)
{
	EtreeModel *e= ETREE_MODEL(tree_model);

	g_return_val_if_fail(IS_ETREE_MODEL(tree_model), -1);
	if ( iter == NULL )
		return e->num_pkts;

	g_return_val_if_fail(iter->stamp == e->stamp, -1);
	g_return_val_if_fail((size_t)iter->user_data < e->num_pkts, -1);

	return 0;
}

static gboolean iter_nth_child(GtkTreeModel *tree_model,
				GtkTreeIter *iter,
				GtkTreeIter *parent,
				gint n)
{
	EtreeModel *e= ETREE_MODEL(tree_model);

	g_return_val_if_fail(IS_ETREE_MODEL(tree_model), FALSE);
	g_return_val_if_fail((size_t)iter->user_data < e->num_pkts, FALSE);

	if ( parent )
		return FALSE;

	iter->stamp = e->stamp;
	iter->user_data = (gpointer)n;

	return TRUE;
}

static gboolean iter_parent(GtkTreeModel *tree_model,
				GtkTreeIter *iter,
				GtkTreeIter *child)
{
	return FALSE;
}

static void etree_tree_model_init(GtkTreeModelIface *i)
{
	i->get_flags = get_flags;
	i->get_n_columns = get_n_columns;
	i->get_column_type = get_column_type;
	i->get_iter = get_iter;
	i->get_path = get_path;
	i->get_value = get_value;
	i->iter_next = iter_next;
	i->iter_children = iter_children;
	i->iter_has_child = iter_has_child;
	i->iter_n_children = iter_n_children;
	i->iter_nth_child = iter_nth_child;
	i->iter_parent = iter_parent;
}

GType etree_model_get_type(void)
{
	static GType etree_model_type=0;

	if ( !etree_model_type ) {
		static const GTypeInfo etree_model_info = {
			sizeof(EtreeModelClass),
			NULL,
			NULL,
			(GClassInitFunc) etree_model_class_init,
			NULL,
			NULL,
			sizeof(EtreeModel),
			0,
			(GInstanceInitFunc) etree_model_init,
		};
		static const GInterfaceInfo tree_model_info = {
			(GInterfaceInitFunc) etree_tree_model_init,
			NULL,
			NULL,
		};

		etree_model_type = g_type_register_static(
					G_TYPE_OBJECT,
					"EtreeModel",
					&etree_model_info,
					0);
		g_type_add_interface_static(etree_model_type,
					GTK_TYPE_TREE_MODEL,
					&tree_model_info);

		icon=load_pixmap("/usr/share/pixmaps/"
			"document-icons/i-chardev.png");

	}

	return etree_model_type;
}

EtreeModel *etree_model_new(void)
{
	EtreeModel *ret;

	ret = g_object_new(ETREE_MODEL_TYPE, NULL);

	return ret;
}

void etree_model_close(EtreeModel *e)
{
	GtkTreePath *path;

	g_return_if_fail(IS_ETREE_MODEL(e));

	while( e->num_pkts ) {
		path = gtk_tree_path_new();
		gtk_tree_path_append_index(path, --e->num_pkts);
		gtk_tree_model_row_deleted(GTK_TREE_MODEL(e), path);
		gtk_tree_path_free(path);
	}

	if ( e->efile )
		elog_free(e->efile);
}

int etree_model_get(EtreeModel *e, GtkTreeIter *iter,
			proc_elogfn cb, gpointer user_data)
{
	g_return_val_if_fail(IS_ETREE_MODEL(e), -1);
	g_return_val_if_fail(e->stamp == iter->stamp, -1);
	g_return_val_if_fail((size_t)iter->user_data < e->num_pkts, -1);

	elog_set_alert_fn(e->efile, cb);
	elog_pkt(e->efile, e->ofs[(size_t)iter->user_data], user_data);
}

static pkt_insert(struct elog_pkt *p, EtreeModel *e)
{
	GtkTreePath *path;
	GtkTreeIter iter;

	iter.stamp = e->stamp;

	path = gtk_tree_path_new();
	gtk_tree_path_append_index(path, e->num_pkts);
	gtk_tree_model_row_inserted(GTK_TREE_MODEL(e), path, &iter);
	gtk_tree_path_free(path);

	e->ofs[e->num_pkts] = p->ofs;

	e->num_pkts++;
}

static pkt_count(struct elog_pkt *p, size_t *cnt)
{
	(*cnt)++;
}

int etree_model_open(EtreeModel *e, int fd)
{
	size_t count=0;

	g_return_val_if_fail(IS_ETREE_MODEL(e), -1);

	/* returns all values to zero */
	etree_model_close(e);

	/* Pick a new iter stamp for each open file */
	e->stamp = g_random_int();

	if ( !(e->efile=elog_new(fd)) )
		return -1;

	/* Initial count */
	elog_set_alert_fn(e->efile, (proc_elogfn)pkt_count);
	elog_run(e->efile, &count);

	/* Zero packets, well, I guess we are done */
	if ( !count )
		return 0;

	/* Create an index->offset map */
	if ( !(e->ofs=calloc(count, sizeof(*e->ofs))) ) {
		int x=errno;
		etree_model_close(e);
		errno=x; /* report correct errors */
		return -1;
	}

	/* Insert all items */
	elog_set_alert_fn(e->efile, (proc_elogfn)pkt_insert);
	elog_run(e->efile, e);

	return 0;
}

void etree_model_set_appbar(EtreeModel *e, GnomeAppBar *a)
{
	g_return_if_fail(IS_ETREE_MODEL(e));
	g_return_if_fail(GNOME_IS_APPBAR(a));

	e->appbar = a;
}
