/*
* This file is part of Firestorm NIDS
* Copyright (c) 2003 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This is the bulk of the GNOME interface code, it's pretty rough
* right now but things should neaten up as we get more of an idea
* how to structure the interface etc. There is noone worth cloning
* after all ;)
*
* TODO:
*  o Allow user to customize column layout
*  o Page based views
*  o Sorting and searching
*  o Grouping
*  o Save as / Save Selection as (convert).
*  o Improve messages view
*  o Put markup tags in details view
*  o Preferences / Properties dialog
*  o Common operations in context menu on packets
*
* Future:
*  o Build indexes, (offset index?)
*  o Tag arbitrary data on to packets (comments, analysis)
*  o Store configuration in GConf
*  o Saved queries and canned queries
*  o Implement plugin/object tree (maybe)
*  o Compressed data store for older data
*  o In-depth event database
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <netinet/in.h>

#include <gnome.h>

#include <firestorm.h>
#include <cleanup.h>
#include <args.h>
#include <packet.h>
#include <alert.h>
#include <signature.h>
#include <target.h>
#include <decode.h>
#include <elog.h>
#include <elog_read.h>
#include <g_etree_model.h>

static EtreeModel *etree;
static GtkTextBuffer *console_buf;
static GtkTextBuffer *pkt_buf;
static GnomeAppBar *appbar;
static void *efile;

/* Exit callbacks */
static void quit_cb(GtkMenuItem *menuitem, gpointer user_data)
{
	gtk_main_quit();
}

static void delete_cb(void)
{
	gtk_main_quit();
}

static void hex_dump(char *ptr, unsigned int len, unsigned int llen)
{
	unsigned int i, j;
	unsigned int line;
	unsigned char *tmp=(unsigned char *)ptr;
	char buf[4*llen + 16];
	char *p;

	for(j=0; j<len; ) {
		p=buf;

		if ( j+llen > len ) {
			line=len-j;
		}else{
			line=llen;
		}

		p+=sprintf(p, "%05X : ", j);

		for(i=0; i<llen; i++) {
			if ( i<line ) {
				if ( isprint(tmp[i]) && !(tmp[i]&0x80)) {
					p+=sprintf(p, "%c", tmp[i]);
				}else{
					p+=sprintf(p, ".");
				}
			}else{
				p+=sprintf(p, " ");
			}
		}

		for(i=0; i<line; i++) {
			p+=sprintf(p, " %02X", tmp[i]);
		}

		j+=line;
		tmp+=line;

		p+=sprintf(p, "\n");
		gtk_text_buffer_insert_at_cursor(pkt_buf, buf, -1);
	}
}

static int do_show(struct alert *a, struct packet *pkt, void *priv)
{
	static char tb[32];
	static char ln[256];
	struct tm *rtime;
	int i;

	rtime=localtime(&pkt->time.tv_sec);
	strftime(tb, sizeof(tb), "%Y-%m-%d %H:%M:%S", rtime);

	snprintf(ln, sizeof(ln), "packet: %s.%.6lu len=%u caplen=%u\n",
		tb, pkt->time.tv_usec, pkt->len, pkt->caplen);
	gtk_text_buffer_insert_at_cursor(pkt_buf, ln, -1);

	snprintf(ln, sizeof(ln), "alert: [%s] %s (sid=%u.%u prio=%u)\n",
		a->gen->name, a->alert,
		a->sid, a->rev, a->priority);
	gtk_text_buffer_insert_at_cursor(pkt_buf, ln, -1);

	for(i=0; i<pkt->llen; i++) {
		if ( !pkt->layer[i].proto ) {
			snprintf(ln, sizeof(ln), "data: "
				"Application layer data (%u bytes)\n",
				pkt->end-pkt->layer[i].h.raw);
			gtk_text_buffer_insert_at_cursor(pkt_buf, ln, -1);
			continue;
		}

		if ( pkt->layer[i].proto->print ) {
			size_t len;
			len=snprintf(ln, sizeof(ln), "%s: ", pkt->layer[i].proto->name);
			pkt->layer[i].proto->print(
				&pkt->layer[i], ln+len, sizeof(ln)-len-1);
			strcat(ln, "\n");
		}else{
			snprintf(ln, sizeof(ln), "%s: (no information)\n",
				pkt->layer[i].proto->name);
		}
		gtk_text_buffer_insert_at_cursor(pkt_buf, ln, -1);
	}

	gtk_text_buffer_insert_at_cursor(pkt_buf, "\n", 1);
	hex_dump(pkt->base, pkt->caplen, 16);

	return 1;
}

static int view_fn(struct elog_pkt *p, void *user_data)
{
	return elog_pkt_alert(p, do_show, &p->ofs);
}

static void show_pkt(GtkTreeModel *model,
			GtkTreePath *path,
			GtkTreeIter *iter,
			gpointer data)
{
	gtk_text_buffer_set_text(pkt_buf, "", 0);
	etree_model_get(ETREE_MODEL(model), iter, view_fn, NULL);
}

static void select_pkt(GtkTreeSelection *sel, gpointer user_data)
{
	gtk_tree_selection_selected_foreach(sel, show_pkt, NULL);
}

void mesg(unsigned char code, char *fmt, ...)
{
	va_list va;
	static char buf[1024];
	gint len;

	va_start(va, fmt);
	len=vsnprintf(buf, sizeof(buf)-2, fmt, va);

	if ( code == M_ERR )
		gnome_error_dialog(buf);
	
	strcat(buf, "\n");
	va_end(va);

	/* TODO: Scroll to new text automatically... */
	gtk_text_buffer_insert_at_cursor(console_buf, buf, len+1);
}

static void display_file(const char *fn)
{
	int fd;
	char buf[32];

	if ( (fd=open(fn, O_RDONLY))<0 ) {
		mesg(M_ERR, "Error opening '%s' - %s.",
			fn, get_err());
		return;
	}

	if ( etree_model_open(etree, fd)<0 ) {
		mesg(M_ERR, "Error reading '%s' - %s.",
			fn, errno ? get_err() : "not an elog file");
		return;
	}

	snprintf(buf, sizeof(buf),
		"%u Alerts Loaded",
		etree_model_num_pkts(etree));
	gnome_appbar_push(appbar, buf);

	close(fd);
}

static void open_file(GtkWidget *w, gpointer user_data)
{
	GtkFileSelection *s=GTK_FILE_SELECTION(user_data);

	gchar **sels=gtk_file_selection_get_selections(s);
	gchar *x;

	while( (x=*sels++) ) {
		display_file(x);
	}
}

/* Open a new file */
static void open_cb(GtkMenuItem *menuitem, gpointer user_data)
{
	GtkWidget *fs;

	/* Multiple file selector */
	fs=gtk_file_selection_new("Select snort rule file");
	gtk_file_selection_set_select_multiple(
		GTK_FILE_SELECTION(fs),FALSE);

	gtk_signal_connect(
		GTK_OBJECT(GTK_FILE_SELECTION(fs)->ok_button),
		"clicked", GTK_SIGNAL_FUNC(open_file), fs);

	/* Make sure object is destroyed when done with */
	gtk_signal_connect_object(
		GTK_OBJECT(GTK_FILE_SELECTION(fs)->ok_button),
		"clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
		(gpointer)fs);
	gtk_signal_connect_object(
		GTK_OBJECT(GTK_FILE_SELECTION(fs)->cancel_button),
		"clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
		(gpointer)fs);

	/* Show it */
	gtk_widget_show(fs);
}

void save_cb(GtkMenuItem *menuitem, gpointer user_data){}
void save_as_cb(GtkMenuItem *menuitem, gpointer user_data){}
void cut_cb(GtkMenuItem *menuitem, gpointer user_data){}
void copy_cb(GtkMenuItem *menuitem, gpointer user_data){}
void paste_cb(GtkMenuItem *menuitem, gpointer user_data){}
void properties_cb(GtkMenuItem *menuitem, gpointer user_data){}
void preferences_cb(GtkMenuItem *menuitem, gpointer user_data){}
void about_cb(GtkMenuItem *menuitem, gpointer user_data)
{
	static GtkWidget *about = NULL;
	GdkPixbuf *pixbuf = NULL;
	gchar *translator_credits = "No translations...";
	gchar *authors[] = {
		"Gianni Tedesco <gianni@scaramanga.co.uk>",
		NULL
	};
	gchar *documenters[] = {
		"Gianni Tedesco <gianni@ecsc.co.uk>",
		"John Leach <john@ecsc.co.uk>",
		NULL
	};

	if (about != NULL) {
		gtk_window_present(GTK_WINDOW (about));
		return;
	}

	about = gnome_about_new("Firestorm Analyst Console", VERSION,
			"Copyright \xc2\xa9 2003 Gianni Tedesco",
			"http://www.scaramanga.co.uk/firestorm/",
			(const char **)authors, (const char **)documenters,
			strcmp (translator_credits, "translator_credits") != 0 ?
			translator_credits : NULL, NULL);

	g_signal_connect(GTK_OBJECT (about), "destroy",
			G_CALLBACK (gtk_widget_destroyed), &about);

	gtk_widget_show(about);
}

static GnomeUIInfo file1_menu_uiinfo[] = {
	GNOMEUIINFO_MENU_OPEN_ITEM(open_cb, NULL),
	GNOMEUIINFO_MENU_SAVE_ITEM(save_cb, NULL),
	GNOMEUIINFO_MENU_SAVE_AS_ITEM(save_as_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_EXIT_ITEM(quit_cb, NULL),
	GNOMEUIINFO_END
};

static GnomeUIInfo edit1_menu_uiinfo[] = {
	GNOMEUIINFO_MENU_CUT_ITEM(cut_cb, NULL),
	GNOMEUIINFO_MENU_COPY_ITEM(copy_cb, NULL),
	GNOMEUIINFO_MENU_PASTE_ITEM(paste_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_PROPERTIES_ITEM(properties_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_PREFERENCES_ITEM(preferences_cb, NULL),
	GNOMEUIINFO_END
};

static GnomeUIInfo view1_menu_uiinfo[] = {
	GNOMEUIINFO_END
};

static GnomeUIInfo help1_menu_uiinfo[] = {
	GNOMEUIINFO_HELP("gnome-firestorm-console"),
	GNOMEUIINFO_MENU_ABOUT_ITEM(about_cb, NULL),
	GNOMEUIINFO_END
};

static GnomeUIInfo menubar1_uiinfo[] = {
	GNOMEUIINFO_MENU_FILE_TREE(file1_menu_uiinfo),
	GNOMEUIINFO_MENU_EDIT_TREE(edit1_menu_uiinfo),
	GNOMEUIINFO_MENU_VIEW_TREE(view1_menu_uiinfo),
	GNOMEUIINFO_MENU_HELP_TREE(help1_menu_uiinfo),
	GNOMEUIINFO_END
};

static GtkWidget *pkt_tree_init(void)
{
	GtkTreeView *tree;
	GtkTreeViewColumn *col;
	GtkCellRenderer *cell;
	GtkWidget *scroll;

	scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

	etree=etree_model_new();

       	tree=GTK_TREE_VIEW(gtk_tree_view_new_with_model(GTK_TREE_MODEL(etree)));
	gtk_tree_view_set_rules_hint(tree, TRUE);
	g_signal_connect(
		G_OBJECT(gtk_tree_view_get_selection(tree)), "changed",
		G_CALLBACK(select_pkt), NULL);

	/* Alert column */
	col=gtk_tree_view_column_new();
	cell=gtk_cell_renderer_pixbuf_new();
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "pixbuf", ETREE_ICON, NULL);
	cell=gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", ETREE_PRIO, NULL);
	gtk_tree_view_column_set_title(col, "Priority");
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_append_column(tree, col);

	col=gtk_tree_view_column_new();
	cell=gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", ETREE_TIME, NULL);
	gtk_tree_view_column_set_title(col, "Time");
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_append_column(tree, col);

	col=gtk_tree_view_column_new();
	cell=gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", ETREE_SRC, NULL);
	gtk_tree_view_column_set_title(col, "Source");
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_append_column(tree, col);

	col=gtk_tree_view_column_new();
	cell=gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", ETREE_DST, NULL);
	gtk_tree_view_column_set_title(col, "Destination");
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_append_column(tree, col);

	col=gtk_tree_view_column_new();
	cell=gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", ETREE_GEN, NULL);
	gtk_tree_view_column_set_title(col, "Alert");
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_append_column(tree, col);

	/* Description column */
	col=gtk_tree_view_column_new();
	cell=gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", ETREE_MSG, NULL);
	gtk_tree_view_column_set_title(col, "Message");
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_append_column(tree, col);

	gtk_container_add(GTK_CONTAINER(scroll), GTK_WIDGET(tree));

	return scroll;
}

static GtkWidget *console_new(void)
{
	GtkWidget *scroll;
	GtkWidget *text;

	scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	console_buf = gtk_text_buffer_new(NULL);
	text = gtk_text_view_new_with_buffer(console_buf);
	gtk_text_view_set_editable(GTK_TEXT_VIEW(text), FALSE);
	gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(text), FALSE);

	gtk_container_add(GTK_CONTAINER(scroll), text);

	return scroll;
}

static GtkWidget *pkt_view_new(void)
{
	GtkWidget *scroll;
	GtkWidget *text;
	PangoFontDescription *font_desc;

	scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	pkt_buf = gtk_text_buffer_new(NULL);
	text = gtk_text_view_new_with_buffer(pkt_buf);
	gtk_text_view_set_editable(GTK_TEXT_VIEW(text), FALSE);
	gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(text), FALSE);

	font_desc = pango_font_description_from_string("Lucida Console 10");
	if ( !font_desc )
		font_desc = pango_font_description_from_string("monospace 10");
	gtk_widget_modify_font(text, font_desc);
	pango_font_description_free(font_desc);

	gtk_container_add(GTK_CONTAINER(scroll), text);

	return scroll;
}

GtkWidget *create_mainwin(void)
{
	GtkWidget *mainwin;
	GtkWidget *toolbar;
	GtkWidget *vpane;
	GtkWidget *tabs;
	GtkWidget *main_box;

	/* Create main window */
	mainwin = gnome_app_new("gnome-firestorm-console",
		"Firestorm Analyst Console");
	gtk_signal_connect(GTK_OBJECT(mainwin),
		"delete-event", delete_cb, NULL);
	gtk_window_set_default_size(GTK_WINDOW(mainwin), 800, 600);

	/* Stick the menus in */
	gnome_app_create_menus(GNOME_APP(mainwin), menubar1_uiinfo);

	/* Add the toolbar  */
	toolbar = gtk_toolbar_new();
	gnome_app_add_toolbar(GNOME_APP(mainwin), GTK_TOOLBAR(toolbar), "toolbar",
		BONOBO_DOCK_ITEM_BEH_EXCLUSIVE, BONOBO_DOCK_TOP, 1, 0, 0);
	gtk_container_set_border_width(GTK_CONTAINER(toolbar), 1);
	gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_BOTH);

	/* The three toolbar buttons */
	gtk_toolbar_insert_stock(GTK_TOOLBAR(toolbar),
		"gtk-open", NULL, NULL,
		GTK_SIGNAL_FUNC(open_cb), NULL, -1);
	gtk_toolbar_insert_stock(GTK_TOOLBAR(toolbar),
		"gtk-save", NULL, NULL,
		GTK_SIGNAL_FUNC(NULL), NULL, -1);
	gtk_toolbar_insert_stock(GTK_TOOLBAR(toolbar),
		"gtk-find", NULL, NULL,
		GTK_SIGNAL_FUNC(NULL), NULL, -1);

	vpane = gtk_vpaned_new();

	/* Packet view */
	main_box = pkt_tree_init();

	/* Tab sheet */
	tabs = gtk_notebook_new();
	gtk_notebook_append_page(GTK_NOTEBOOK(tabs),
		pkt_view_new(),
		gtk_label_new("Details"));
	gtk_notebook_append_page(GTK_NOTEBOOK(tabs),
		console_new(),
		gtk_label_new("Messages"));

	gtk_paned_add1(GTK_PANED(vpane), main_box);
	GTK_PANED(vpane)->child1_resize=TRUE;
	GTK_PANED(vpane)->child1_shrink=TRUE;

	gtk_paned_add2(GTK_PANED(vpane), tabs);
	GTK_PANED(vpane)->child2_resize=FALSE;
	GTK_PANED(vpane)->child2_shrink=FALSE;

	gnome_app_set_contents(GNOME_APP(mainwin), vpane);

	/* Status bar */
	appbar = (GnomeAppBar *)gnome_appbar_new(TRUE, TRUE,
			GNOME_PREFERENCES_USER);
	gnome_app_set_statusbar(GNOME_APP(mainwin), GTK_WIDGET(appbar));

	/* Tool tips */
	gnome_app_install_menu_hints(GNOME_APP(mainwin), menubar1_uiinfo);

	etree_model_set_appbar(etree, appbar);

	return mainwin;
}
