/*
* This file is part of tcpidx
* Copyright (c) 2003 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*
* avl tree maintenance functions taken from avl-0.4, original copyright
* notice is reproduced below:
*
* (C) 2000 Daniel Nagy, Budapest University of Technology and Economics
* Released under GNU General Public License (GPL) version 2
*/

#include <stdlib.h>
#include <string.h>

#include <mpool.h>
#include <fbuf.h>
#include <idx.h>
#include <i_ordered.h>

struct avl {
	struct avl *left, *right;
	idx_key_t key;
	int balance;
	struct mpool vals;
	idx_ofs_t num_vals;
};

static struct avl *cur;

/* avl_swl: 
 *
 * AVL tree maintenance function
*/
static inline void avl_swl(struct avl **r)
{
	struct avl *a=*r;
	struct avl *b=a->right;
	*r=b;
	a->right=b->left;
	b->left=a;
}

/* avl_swr: 
 *
 * AVL tree maintenance function
*/
static inline void avl_swr(struct avl **r)
{
	struct avl *a=*r;
	struct avl *b=a->left;
	*r=b;
	a->left=b->right;
	b->right=a;
}

/* avl_balance: 
 *
 * AVL tree maintenance function
*/
static inline void avl_balance(struct avl *r)
{
	switch ( r->balance ) {
	case -1:
		r->left->balance=0;
		r->right->balance=1;
		break;
	case 1:
		r->left->balance=-1;
		r->right->balance=0;
		break;
	case 0:
		r->left->balance=0;
		r->right->balance=0;
	}
	r->balance=0;
}

static inline int hdr_count(struct mpool *m)
{
	int ret=0;
	struct mpool_hdr *h;

	for(h=m->slabs; h; h=h->next)
		ret++;

	return ret;
}

static void reverse_mpool(struct mpool *m)
{
	int cnt=hdr_count(m);
	struct mpool_hdr *a[cnt];
	struct mpool_hdr *h;
	int i=0;

	for(h=m->slabs; h; h=h->next)
		a[i++] = h;

	m->slabs=NULL;

	for(i=0; i<cnt; i++) {
		a[i]->next = m->slabs;
		m->slabs = a[i];
	}
}

/* avl_insert_key:
 * @r: double-indirect pointer to the root #avl node
 * @n: the node to insert
 *
 * Insert an AVL node in to the tree, sets the global variable cur to be
 * a pointer to the unique node to use. If there is already a node for
 * that key in the tree @n will not be used.
 *
 * Return value: for recursion only, to be ignored.
 */
static int avl_insert_key(struct avl **r, struct avl *n)
{
	struct avl *t=*r;

	if ( n->key < t->key ) {
		if ( t->left ) {
			if ( avl_insert_key(&t->left, n) ) {
				switch ( t->left->balance-- ) {
				case 1: return 0;
				case 0: return 1;
				}
				if ( t->left->balance<0 ) {
					/* simple rotate */
					avl_swr(r);
					(*r)->balance=0;
					(*r)->right->balance=0;
				}else{
					/* double rotate */
					avl_swl(&t->left);
					avl_swr(r);
					avl_balance(*r);
				}
			}
			return 0;
		}else{
			t->left=n;
			if ( t->balance-- ) return 0;
			return 1;
		}
	}

	if ( n->key > t->key ) {
		if ( t->right ) {
			if ( avl_insert_key(&t->right, n) ) {
				switch ( t->right->balance-- ) {
				case 1: return 0;
				case 0: return 1;
				}
				if ( t->right->balance<0 ) {
					/* simple rotate */
					avl_swl(r);
					(*r)->balance=0;
					(*r)->left->balance=0;
				}else{
					/* double rotate */
					avl_swr(&t->right);
					avl_swl(r);
					avl_balance(*r);
				}
			}
			return 0;
		}else{
			t->right=n;
			if ( t->balance-- ) return 0;
			return 1;
		}
	}

	if ( n->key==t->key )
		cur=t;

	return 0;
}

/* i_ordered_new:
 * @db: an #i_ordered structure to use
 *
 * Prepares an #i_ordered structure for inserting records in to.
 *
 * Return value: zero on success, less than zero for error.
 */
int i_ordered_new(struct i_ordered *db)
{
	/* Create memory allocator for AVL tree nodes */
	if ( mpool_new(&db->avl_nodes, sizeof(struct avl), 0)<0 )
		return -1;

	db->cache=NULL;
	db->root=NULL;
	db->num_keys=0;
	db->num_vals=0;

	return 0;
}

/* i_ordered_insert:
 * @db: a valid #i_ordered database
 * @key: key to use
 * @val: offset in file being indexed
 *
 * Inserts an item in to the index.
 *
 * Return value: zero on success, less than zero for error.
 */
int i_ordered_insert(struct i_ordered *db,
			idx_key_t key,
			idx_val_t val)
{
	struct avl *n;
	int new_key=0;
	idx_val_t *v;

	if ( db->cache ) {
		n=db->cache;
	}else{
		n=db->cache=mpool_alloc(&db->avl_nodes);
		if ( n==NULL )
			return -1;
	}

	n->left=NULL;
	n->right=NULL;
	n->key=key;
	n->balance=0;
	n->num_vals=0;
	cur=n;

	if ( db->root ) {
		avl_insert_key(&db->root, n);
		if ( cur==n )
			new_key=1;
	}else{
		db->root=n;
		new_key=1;
	}

	if ( new_key ) {
		if ( mpool_new(&cur->vals, sizeof(idx_val_t), 0)<0 )
			return -1;

		db->num_keys++;
		cur->num_vals=0;
		db->cache=NULL;
	}

	/* Add the value */
	if ( !(v=mpool_alloc(&cur->vals)) )
		return -1;

	*v=val;
	db->num_vals++;
	cur->num_vals++;

	return 0;
}

static void i_ordered_free_vals(struct avl *n)
{
	if ( n->left )
		i_ordered_free_vals(n->left);

	mpool_free(&n->vals);

	if ( n->right )
		i_ordered_free_vals(n->right);
}

/* i_ordered_free:
 * @db: a valid #i_ordered database
 *
 * Deletes all data currently associated with the database. The 
 */
void i_ordered_free(struct i_ordered *db)
{
	if ( db->root )
		i_ordered_free_vals(db->root);

	mpool_free(&db->avl_nodes);
	db->num_keys=0;
	db->num_vals=0;
	db->cache=NULL;
}

static idx_ofs_t cur_ofs;

static int i_ordered_write_key(struct avl *n, struct fbuf *o)
{
	if ( n->left ) {
		if ( i_ordered_write_key(n->left, o)<0 )
			return -1;
	}

	do {
		struct idx_node dnode;

		dnode.key = n->key;
		dnode.ofs = cur_ofs;
		dnode.num = n->num_vals;

		cur_ofs += sizeof(idx_val_t) * n->num_vals;

		if ( fbuf_write(o, &dnode, sizeof(dnode))<0 )
			return -1;
	}while(0);

	if ( n->right ) {
		if ( i_ordered_write_key(n->right, o)<0 )
			return -1;
	}

	return 0;
}

static int i_ordered_write_val(struct avl *n, struct fbuf *o)
{
	if ( n->left ) {
		if ( i_ordered_write_val(n->left, o)<0 )
			return -1;
	}

	do {
		struct mpool_hdr *h;
		int tlen=0;

		reverse_mpool(&n->vals);

		for(h=n->vals.slabs; h; h=h->next) {
			size_t len;

			/* FIXME: once we add alignment and/or coloring
			 * this will fail badly */
			len = (char *)h->next_obj - (char *)h;
			if ( len > n->vals.slab_size )
				len -= n->vals.obj_size;

			len -= sizeof(*h);

			tlen += len;

			if ( fbuf_write(o, h->data, len)<0 )
				return -1;
		}

		/* to catch errors iterating the mpool */
		if ( tlen != n->num_vals*sizeof(idx_val_t) )
			return -1;
	}while(0);

	if ( n->right ) {
		if ( i_ordered_write_val(n->right, o)<0 )
			return -1;
	}

	return 0;
}

/* i_ordered_write:
 * @db: a valid #i_ordered database
 * @o: an output file buffer
 *
 * Writes the database out to the file
 *
 * Return value: zero on success, less than zero for error.
 */
int i_ordered_write(struct i_ordered *db, struct fbuf *o)
{
	cur_ofs = db->num_keys * sizeof(struct idx_node);
	
	if ( !db->root )
		return 0;

	if ( i_ordered_write_key(db->root, o) < 0 )
		return -1;

	if ( i_ordered_write_val(db->root, o) < 0 )
		return -1;

	return 0;
}
